
/* Copyright (C) 2002-2008 Free Software Foundation, Inc.
  Contributed by Andy Vaught

  This file is part of g95.

  G95 is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  G95 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g95; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  In addition to the permissions in the GNU General Public License, the
  Free Software Foundation gives you unlimited permission to link the
  compiled version of this file into combinations with other programs,
  and to distribute those combinations without any restriction coming
  from the use of this file.  (The General Public License restrictions
  do apply in other respects; for example, they cover modification of
  the file, and distribution when not linked into a combined executable.)
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stddef.h>
#include <signal.h>

#include "runtime.h"


#if HAVE_DL_MALLOC
#include <malloc.h>
#endif


static st_parameter iobase;
st_parameter *ioparm = NULL;
global_t globals;

int init_flag=0;

#define string_len prefix(string_len)
#define library_return prefix(library_return)

G95_DINT string_len, junk_stat, library_return;

char **f__xargv;
int f__xargc;


/* If a --g95 is present, g95_argv will be non-NULL, even if g95_argc
 * is zero.  A non-NULL g95_argv and a zero g95_argc causes
 * environment settings to be dumped.  Unlink argv[0], g95_argv[0] is
 * really the first argument and g95_argc is really the number of
 * library arguments. */

static char **g95_argv;
static int g95_argc;



void get_ioparm(void) {
st_parameter *p;

    p = (ioparm == NULL)
	? &iobase
	: temp_alloc(sizeof(st_parameter));

    memset(p, '\0', sizeof(st_parameter));

    p->prev = ioparm;
    p->unit_save = current_unit;
    p->file_save = filename;
    p->line_save = line;

    ioparm = p;
}


/* library_start()-- Called when an I/O call starts.  */

void library_start(void) {

    if (!init_flag)
	g95_runtime_start(0, NULL);

    if (ioparm->iostat != NULL)
	set_integer(ERROR_OK, ioparm->iostat, ioparm->iostat_kind);

    ioparm->library_rc = LIBRARY_OK;

#if HAVE_RESUME
    if (ioparm == &iobase) {
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, SIGQUIT);
	sigaddset(&set, SIGALRM);
	sigprocmask(SIG_BLOCK, &set, NULL);
    }
#endif
}



/* library_end()-- Called when a library call is complete in order to
 * clean up for the next call. */

void library_end(void) {
st_parameter *p;

    library_return = ioparm->library_rc;

    if (ioparm == &iobase) {
	current_unit = NULL;
	filename = NULL;
	line = 0;

	memset(&ioparm, '\0', sizeof(ioparm));
	ioparm = NULL;

    } else {
	current_unit = ioparm->unit_save;
	filename = ioparm->file_save;
	line = ioparm->line_save;

	p = ioparm->prev;
	temp_free((void *) &ioparm);
	ioparm = p;
    }

#if HAVE_RESUME
    if (ioparm == NULL) {
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, SIGQUIT);
	sigaddset(&set, SIGALRM);
	sigprocmask(SIG_UNBLOCK, &set, NULL);
    }
#endif
}



/* xorshift128()-- Internal random number generator used for various
 * things.  Developed by George Marsaglia, it has a period of 2^128-1
 * and passes all of his DIEHARD tests for random numbers. */

int xorshift128() {
static unsigned long x=123456789, y=362436069, z=521288629, w=88675123;
unsigned long t;

    t = x ^ (x << 20) ^ y ^ (y >> 11) ^ z ^ (z << 27) ^ w ^ (w << 6);

    x = y;
    y = z;
    z = w;
    w = t;

    return w;
}



/* init_arglist()-- Split the command line arguments into g95-specific
 * arguments and arguments to be passed to the program.  We start out
 * looking for program arguments.  When we see a --g95, that starts
 * g95 library arguments.  When we see '--', that means that the rest
 * of the arguments are for the program.  */

static void init_arglist(int argc, char *argv[]) {
int i, j, k, g95_pos, dash_pos;

    f__xargv = argv;

    g95_pos  = -1;
    dash_pos = -1;

    for(i=1; i<argc; i++)
	if (strcmp(argv[i], "--g95") == 0)
	    g95_pos = i;

	else if (strcmp(argv[i], "--") == 0) {
	    dash_pos = i;
	    break;
	}

    /* Cases */

    if (g95_pos != -1 && dash_pos != -1) {
	/* We have both g95 and program arguments.  Paste the trailing
	 * program arguments to the leading, moving the g95 arguments
	 * to the end of the existing list.  We transform:
	 *      A1 A2 --g95 G1 G2 -- A3 A4
	 * to   A1 A2 A3 A4 NULL G1 G2 */

	i = g95_pos;
	j = dash_pos + 1;
	k = g95_pos + 1;

	while(j < argc) {
	    argv[i] = argv[j];
	    memmove(&argv[k+1], &argv[k],
		    (dash_pos - g95_pos - 1)*sizeof(argv[0]));

	    i++; j++; k++;
	}

	argv[i]  = NULL;
	f__xargc = g95_pos + argc - dash_pos - 1;

	g95_argc = dash_pos - g95_pos - 1;
	g95_argv = &argv[k];

    } else if (g95_pos != -1 && dash_pos == -1) {
	/* In this case, we might have some program arguments followed
	 * by g95 arguments.  The list is almost the way we want it.
	 * We transform:
	 *     A1 A2 --g95 G1 G2
	 * to  A1 A2 NULL G1 G2   */

	argv[g95_pos] = NULL;
	f__xargc = g95_pos;

	g95_argv = &argv[g95_pos + 1];
	g95_argc = argc - g95_pos - 1;

    } else if (dash_pos != -1) {
	/* In this case we have -- by itself, presumably to escape a
	 * --g95.  We eliminate the dash argument. */

	memmove(&argv[dash_pos], &argv[dash_pos+1],
		(argc - dash_pos) * sizeof(argv[0]));

	f__xargc = argc - 1;
	g95_argv = NULL;
	g95_argc = 0;

    } else {
	/* The remaining case has neither -- or --g95.  Pass the
	 * command line as-is, without massaging. */

	f__xargc = argc;
	g95_argv = NULL;
	g95_argc = 0;
    }

#if 0
    printf("Regular: argc=%d   ", f__xargc);
    for(i=1; i<f__xargc; i++)
	printf("%s ", f__xargv[i]);

    putchar('\n');

    if (g95_argv != NULL) {
	printf("Library: argc=%d   ", g95_argc);
	
	for(i=0; i<g95_argc; i++)
	    printf("%s ", g95_argv[i]);

	putchar('\n');
    }
#endif

}



/* g95_runtime_start()-- Get things ready to run */

void g95_runtime_start(int argc, char *argv[]) {
enum { CA_NONE, CA_MASTER, CA_HOST } ca_mode;
int i, images;
char *p;

    init_variables();   /* Must be first */

#if HAVE_DL_MALLOC
    mallopt(M_MXFAST,         options.mem_mxfast);
    mallopt(M_TRIM_THRESHOLD, options.mem_trim_threshold);
    mallopt(M_TOP_PAD,        options.mem_top_pad);
#endif

    init_units();
    init_fpu();
    init_fpu_trap();
    init_random_seed(1);
    init_arglist(argc, argv);

    /* Check runtime variables */

    images = -1;
    ca_mode = CA_NONE;

    if (g95_argv != NULL) {
	if (g95_argc == 0)
	    show_variables();

	for(i=0; i<g95_argc; i++) {
	    p = g95_argv[i];

	    if (strncmp(p, "images=", 7) == 0)
		images = atoi(p + 7);

	    else if (strcmp(p, "master") == 0)
		ca_mode = CA_MASTER;

	    else if (strcmp(p, "host") == 0)
		ca_mode = CA_HOST;

	    else
		st_printf("Unknown library argument '%s'\n", g95_argv[i]);
	}
    }

    switch(ca_mode) {
    case CA_NONE:
	if (images < 1)
	    images = 1;

	init_coarray(images);        /* SMP mode */
	break;

    case CA_MASTER:
	if (images < 1)
	    runtime_error("Images must be specifed in master mode\n");

	_exit(master_init(images, 0));


    case CA_HOST:
	if (images < 1)
	    runtime_error("Images must be specifed in master mode\n");

	host_init(images, 0, NULL);
	break;
    }

    init_memory();

    init_flag = 1;
}



/* g95_runtime_stop()-- Finish up runtime things */

void g95_runtime_stop(void) {

    if (options.view_exceptions)
	show_exceptions();

    close_units();
}
